/*
    Drools5 Integration Helper
    Copyright (C) 2009  Mathieu Boretti mathieu.boretti@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.boretti.drools.integration.drools5.implementation;

import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.boretti.drools.integration.drools5.DroolsInterface;
import org.boretti.drools.integration.drools5.DroolsServiceType;
import org.boretti.drools.integration.drools5.annotations.DroolsService;
import org.boretti.drools.integration.drools5.exceptions.NotAnInterfaceDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedClassDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedInterfaceDroolsError;
import org.drools.RuleBase;

public class DroolsProviderImplementation {
	
	private final Logger logger = Logger.getLogger(this.getClass());
	
	private final Map<String,DroolsInterfaceImplementation<?>> cache = new HashMap<String,DroolsInterfaceImplementation<?>>();
	
	private DroolsProviderImplementation() {}
	
	private static final DroolsProviderImplementation instance = new DroolsProviderImplementation();
	
	public static final DroolsProviderImplementation getInstance() {
		return instance;
	}
	
	@SuppressWarnings("unchecked")
	public <T> T getService(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsService ds = getAnnotationFromInterface(clazz);
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		DroolsImplementationHelper.validNullability(dst,"Invalid type parameter of annotation - can't be null");
		DroolsInterfaceImplementation<T> inter=DroolsImplementationHelper.getDroolsImplementation(dst,resource,clazz);
		if (logger.isDebugEnabled()) logger.debug("Creating proxy using "+clazz+" from "+dst+" using resource "+resource);
		return (T) Proxy.newProxyInstance(clazz.getClassLoader(),new Class[] {clazz,DroolsInterface.class},new DroolsProxy<T>(clazz,inter));
	}
	
	/**
	 * 
	 * @param <T>
	 * @param clazz
	 * @param resourceName
	 * @return
	 * @since 1.2.0
	 */
	@SuppressWarnings("unchecked")
	public <T> T getService(Class<T> clazz,String resourceName) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsImplementationHelper.validNullability(resourceName,"External resourceName can't be null");
		DroolsService ds = getAnnotationFromInterface(clazz);
		DroolsServiceType dst=ds.type();
		DroolsImplementationHelper.validNullability(dst,"Invalid type parameter of annotation - can't be null");
		DroolsInterfaceImplementation<T> inter=DroolsImplementationHelper.getDroolsImplementation(dst,resourceName,clazz);
		if (logger.isDebugEnabled()) logger.debug("Creating proxy using "+clazz+" from "+dst+" using resource "+resourceName);
		return (T) Proxy.newProxyInstance(clazz.getClassLoader(),new Class[] {clazz,DroolsInterface.class},new DroolsProxy<T>(clazz,inter));
	}
	
	/**
	 * 
	 * @param <T>
	 * @param clazz
	 * @param resourceName
	 * @param serviceType
	 * @return
	 * @since 1.2.0
	 */
	@SuppressWarnings("unchecked")
	public <T> T getService(Class<T> clazz,String resourceName,DroolsServiceType serviceType) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsImplementationHelper.validNullability(resourceName,"External resourceName can't be null");
		DroolsImplementationHelper.validNullability(serviceType,"External serviceType can't be null");
		DroolsInterfaceImplementation<T> inter=DroolsImplementationHelper.getDroolsImplementation(serviceType,resourceName,clazz);
		if (logger.isDebugEnabled()) logger.debug("Creating proxy using "+clazz+" from "+serviceType+" using resource "+resourceName);
		return (T) Proxy.newProxyInstance(clazz.getClassLoader(),new Class[] {clazz,DroolsInterface.class},new DroolsProxy<T>(clazz,inter));
	}
	
	/**
	 * 
	 * @param <T>
	 * @param clazz
	 * @return
	 * @since 1.2.0
	 */
	private <T> DroolsService getAnnotationFromInterface(Class<T> clazz) {
		DroolsImplementationHelper.validNullability(clazz,"Invalid interface - can't be null");
		if (!clazz.isInterface()) {
			String msg = "Not an interface "+clazz;
			logger.error(msg);
			throw new NotAnInterfaceDroolsError(msg);
		}
		DroolsService ds = DroolsImplementationHelper.getAnnotation(clazz);
		if(ds==null) {
			String msg = "Invalid interface (missing annotation) "+clazz;
			logger.error(msg);
			throw new NotAnnotatedInterfaceDroolsError(msg);
		}
		return ds;
	}
	
	/**
	 * 
	 * @param <T>
	 * @param clazz
	 * @return
	 * @since 1.2.0
	 */
	private <T> DroolsService getAnnotationFromClass(Class<T> clazz) {
		DroolsImplementationHelper.validNullability(clazz,"Invalid class - can't be null");
		DroolsService ds = DroolsImplementationHelper.getAnnotation(clazz);
		if(ds==null) {
			if (clazz.isInterface()) {
				String msg = "Invalid interface (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedInterfaceDroolsError(msg);
			} else {
				String msg = "Invalid class (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedClassDroolsError(msg);
			}
		}
		return ds;
	}
	
	/**
	 * 
	 * @param clazz
	 * @return
	 * @since 1.0.0
	 */
	public <T> RuleBase getRuleBase(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsService ds = getAnnotationFromClass(clazz);
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		DroolsImplementationHelper.validNullability(dst,"Invalid type parameter of annotation - can't be null");
		return DroolsImplementationHelper.getDroolsImplementation(dst,resource,clazz).getRuleBase();
	}
	
	/**
	 * 
	 * @since 1.2.0
	 */
	public <T> RuleBase getRuleBaseOverride(Class<T> clazz,String resourceName) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsService ds = getAnnotationFromClass(clazz);
		DroolsServiceType dst=ds.type();
		DroolsImplementationHelper.validNullability(dst,"Invalid type parameter of annotation - can't be null");
		return DroolsImplementationHelper.getDroolsImplementation(dst,resourceName,clazz).getRuleBase();
	}
	
	/**
	 * 
	 * @since 1.2.0
	 */
	public <T> RuleBase getRuleBaseOverride(Class<T> clazz,DroolsServiceType resourceType) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsService ds = getAnnotationFromClass(clazz);
		String resource = ds.resourceName();
		DroolsImplementationHelper.validNullability(resourceType,"Invalid type parameter of annotation - can't be null");
		return DroolsImplementationHelper.getDroolsImplementation(resourceType,resource,clazz).getRuleBase();
	}
	
	/**
	 * 
	 * @since 1.2.0
	 */
	public <T> RuleBase getRuleBaseOverride(Class<T> clazz,String resourceName,DroolsServiceType resourceType) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		DroolsImplementationHelper.validNullability(resourceType,"Invalid type parameter of annotation - can't be null");
		return DroolsImplementationHelper.getDroolsImplementation(resourceType,resourceName,clazz).getRuleBase();
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @return
	 * @since 1.1.0
	 */
	@SuppressWarnings("unchecked")
	public <T> RuleBase getRuleBase(Class<T> clazz,DroolsServiceType type,String resourceName) {
		DroolsImplementationHelper.validNullability(clazz,"Class can't be null");
		DroolsImplementationHelper.validNullability(type,"Type can't be null");
		DroolsImplementationHelper.validNullability(resourceName,"ResourceName can't be null");
		String key = clazz.getName()+":"+type+":"+resourceName;
		synchronized(cache) {
			DroolsInterfaceImplementation<?> element = cache.get(key);
			if (element!=null) {
				return element.getRuleBase();
			}
			element = DroolsImplementationHelper.getDroolsImplementation(type,resourceName,clazz);
			cache.put(key,element);
			return element.getRuleBase();
		}
	}

	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @param error
	 * @param reference
	 * @param args
	 * @since 1.1.0
	 */
	public <T> void runPreCondition(Class<T> clazz,
			DroolsServiceType type,
			String resourceName,
			Class<? extends Error> error,
			Object reference,
			Object args[]) {
		DroolsImplementationHelper.runCondition(this,clazz, type, resourceName, error, reference, args);
	}
	
	/**
	 * 
	 * @param clazz
	 * @param type
	 * @param resourceName
	 * @param error
	 * @param reference
	 * @param args
	 * @since 1.1.0
	 */
	public <T> void runPostCondition(Class<T> clazz,
			DroolsServiceType type,
			String resourceName,
			Class<? extends Error> error,
			Object reference,
			Object args[]) {
		DroolsImplementationHelper.runCondition(this,clazz, type, resourceName, error, reference, args);
	}
	
	/**
	 * 
	 * @since 1.1.0
	 */
	static final String FIELD_NAME_LOGGER = "droolsLogger";

}
